/*
    Copyright (C) 2006 by OXIM TEAM

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifdef HAVE_CONFIG_H
#  include "config.h"
#endif

#include <stdlib.h>
#include <string.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/xpm.h>
#include <X11/Xos.h>
#include <X11/extensions/shape.h>
#include "oximtool.h"
#include "gui.h"
#include "oxim.h"

#include "images/oxim_tray.xpm"
int check_tray_manager(void);

static int  tray_init = False;
static char **icon = NULL;
static Atom selection_atom = None;
static Atom manager_atom = None;
static Atom system_tray_opcode_atom = None;
static Atom orientation_atom = None;
static Atom xembed_atom = None;
static Window manager_window = None;
static unsigned int stamp = 0;

#define SYSTEM_TRAY_REQUEST_DOCK    0
#define SYSTEM_TRAY_BEGIN_MESSAGE   1
#define SYSTEM_TRAY_CANCEL_MESSAGE  2

static void send_message(
     Display* dpy, /* display */
     Window win,   /* sender (tray icon window) */
     long message, /* message opcode */
     long data1,   /* message data 1 */
     long data2,   /* message data 2 */
     long data3    /* message data 3 */
)
{
    XEvent ev;
  
    bzero(&ev, sizeof(ev));
    ev.xclient.type = ClientMessage;
    ev.xclient.window = win;
    ev.xclient.message_type = system_tray_opcode_atom;
    ev.xclient.format = 32;
    ev.xclient.data.l[0] = CurrentTime;
    ev.xclient.data.l[1] = message;
    ev.xclient.data.l[2] = data1;
    ev.xclient.data.l[3] = data2;
    ev.xclient.data.l[4] = data3;

//    trap_errors();
    XSendEvent(dpy, manager_window, False, NoEventMask, &ev);
    XSync(dpy, False);
/*
    if (untrap_errors()) {
    }
*/
}

/* 傳送訊息 */
static void gui_send_message(winlist_t *win)
{
    unsigned int timeout = 30000; /* 3 秒 */
    char *msg = "OXIM 輸入法"; /* message */
    int len = strlen(msg);

    stamp++;

    send_message(gui->display, /* display */
     		 manager_window,   /* sender (tray icon window) */
		 SYSTEM_TRAY_BEGIN_MESSAGE, /* message opcode */
		 timeout,   /* message data 1 */
		 len,   /* message data 2 */
		 stamp /* message data 3 */
		);

    while (len > 0)
    {
	XClientMessageEvent ev;

	ev.type = ClientMessage;
	ev.window = manager_window;
	ev.format = 8;
	ev.message_type = XInternAtom(gui->display,
			"_NET_SYSTEM_TRAY_MESSAGE_DATA", False);
	if (len > 20)
	{
	    memcpy(&ev.data, msg, 20);
	    len -= 20;
	    msg += 20;
	}
	else
	{
	    memcpy(&ev.data, msg, len);
	    len = 0;
	}
	XSendEvent(gui->display, manager_window, False, StructureNotifyMask, (XEvent *)&ev);
	XSync(gui->display, False);
    }
}

static void gui_tray_draw(winlist_t *win)
{
    xccore_t *xccore = (xccore_t *)win->data;
    IC *ic = xccore->ic;
    if (win->winmode & WMODE_EXIT || !tray_init)
	return;

    inp_state_t inp_state = (ic) ? ic->imc->inp_state : 0;
    char **disp_icon = oxim_tray;

    Pixmap tray, mask;
    XpmAttributes attributes;
    attributes.valuemask = 0;

    int status = XpmCreatePixmapFromData(gui->display, win->window,
			disp_icon, &tray, &mask, &attributes);
    if (status == 0)
    {
	/* 設定長寬 */
	win->width  = attributes.width;
	win->height = attributes.height;
	/* 重設視窗大小 */
	XResizeWindow(gui->display, win->window, win->width, win->height);
	/* 設定視窗背景 */
//	XSetWindowBackgroundPixmap(gui->display, win->window, tray);
	if (mask)
	{
	    XShapeCombineMask(gui->display, win->window,
			ShapeBounding, 0, 0, mask, ShapeSet);
	}
	XpmFreeAttributes(&attributes);
	XFreePixmap(gui->display, tray);
	if (mask)
	{
	    XFreePixmap(gui->display, mask);
	}
    }
    gui_Draw_Image(win, 0, 0, disp_icon);
}

static void gui_focus_event(winlist_t *win, XEvent *event)
{
    switch (event->type)
    {
	case MapNotify:
	    DebugLog(1, ("MapNotify\n"));
	    break;
	case UnmapNotify:
	    DebugLog(1, ("UnMapNotify\n"));
	    break;
	case ReparentNotify:
	    DebugLog(1, ("ReparentNotify Root(%d), Event_win(%d), Parent_Win(%d)\n", gui->root, event->xreparent.event, event->xreparent.parent));
	    if (event->xreparent.parent == gui->root)
	    {
		DebugLog(1, ("脫離了 Manager Window\n"));
	    }
	    break;
	case ConfigureNotify:
	    DebugLog(1, ("ConfigureNotify\n"));
	    break;
	case MotionNotify: /* Mouse Over */
	    break;
	case EnterNotify: /* Mouse In */
	    DebugLog(1, ("EnterNotify\n"));
	    //gui_send_message(win);
	    break;
	case LeaveNotify: /* Mouse Out */
	    break;
	case Expose:
	    if (event->xexpose.count == 0)
	    {
		gui_tray_draw(win);
	    }
	    break;
	case ButtonPress: /* 按下滑鼠按鍵 */
	    if (event->xbutton.button == Button1)
	    {
		if (gui_menu_actived())
		{
		    gui_hide_menu();
		}
		else
		{
		    gui_show_menu(win);
		}
	    }
	    break;
	case ClientMessage: /* Client message */
	    if (event->xclient.message_type == xembed_atom)
	    {
		DebugLog(1, ("in tray %d\n", event->xclient.data.l[1]));
	    }
	    else
	    {
		DebugLog(1, ("Unknow Client message -> %d\n", event->xclient.message_type));
	    }
	    break;
	case PropertyNotify:
	    DebugLog(1, ("PropertyNotify\n"));
	    break;
	case DestroyNotify:
	    DebugLog(1, ("DestroyNotify\n"));
	    break;
	default:
	    DebugLog(1, ("tray win event %d\n", event->type));
    }
}

/*
    建立 Tray 
*/
void gui_tray_init(void)
{
    winlist_t *win = NULL;

    if (gui->tray_win)
    {
	win = gui->tray_win;
    }
    else
    {
	win = gui_new_win();
    }

    /* 邊框顏色 */
    XSetWindowBorder(gui->display, win->window, gui_Color(WINBOX_COLOR));
    /* 背景顏色 */
    XSetWindowBackground(gui->display, win->window, gui_Color(BORDER_COLOR));

    /* Tray 不需要邊框 */
    XSetWindowBorderWidth(gui->display, win->window, 0);

    win->width  = 22;
    win->height = 20;
    win->pos_x = 0;
    win->pos_y = 0;

    win->win_draw_func = gui_tray_draw;
    win->win_event_func = gui_focus_event;
    XSizeHints *size_hints = XAllocSizeHints();
    size_hints->flags = PMinSize | PMaxSize;
    size_hints->min_width = win->width;
    size_hints->min_height = win->height;
    size_hints->max_width = win->width;
    size_hints->max_height = win->height;
    XSetWMNormalHints(gui->display, win->window, size_hints);
    XFree(size_hints);

    Atom xembed_info = XInternAtom(gui->display, "_XEMBED_INFO", False);
    unsigned int mapping[2];
    mapping[0] = 0; /* ver 0 */
    mapping[1] = 1; /* request mapping */
    XChangeProperty(gui->display, win->window, xembed_info, xembed_info,
			32, PropModeReplace, (unsigned char *)mapping, 2);
    gui->tray_win = win;
}

int check_tray_manager(void)
{
    if (tray_init)
	return True;

    char tray_mgr[32];
    sprintf(tray_mgr, "_NET_SYSTEM_TRAY_S%d", gui->screen);
    selection_atom = XInternAtom(gui->display, tray_mgr, False);
    if (selection_atom == None)
    {
	return False;
    }

    manager_atom = XInternAtom(gui->display, "MANAGER", False);
    xembed_atom = XInternAtom(gui->display, "_XEMBED", False);
    system_tray_opcode_atom = XInternAtom(gui->display, "_NET_SYSTEM_TRAY_OPCODE", False);
    orientation_atom = XInternAtom(gui->display, "_NET_SYSTEM_TRAY_ORIENTATION", False);

    XGrabServer(gui->display);
    manager_window = XGetSelectionOwner(gui->display, selection_atom);

    if (manager_window != None)
    {
	XSelectInput(gui->display, manager_window, StructureNotifyMask|PropertyChangeMask);
	send_message(gui->display, manager_window, SYSTEM_TRAY_REQUEST_DOCK, gui->tray_win->window, 0, 0);
	tray_init = True;
    }
    XUngrabServer(gui->display);
    XFlush(gui->display);

    return tray_init;
}
